import { ExecutionResult, DocumentNode } from 'graphql';
import { ApolloCache } from 'apollo-cache';
import { FragmentMap, DirectiveInfo } from 'apollo-utilities';
import ApolloClient from '../ApolloClient';
import { Initializers, Resolvers, OperationVariables } from './types';
export declare type Resolver = (fieldName: string, rootValue: any, args: any, context: any, info: ExecInfo) => any;
export declare type VariableMap = {
    [name: string]: any;
};
export declare type ResultMapper = (values: {
    [fieldName: string]: any;
}, rootValue: any) => any;
export declare type FragmentMatcher = (rootValue: any, typeCondition: string, context: any) => boolean;
export declare type ExecContext = {
    fragmentMap: FragmentMap;
    context: any;
    variables: VariableMap;
    resultMapper?: ResultMapper;
    fragmentMatcher: FragmentMatcher;
    defaultOperationType?: string | null;
    exportedVariables: Record<string, any>;
};
export declare type ExecInfo = {
    isLeaf: boolean;
    resultKey: string;
    directives: DirectiveInfo;
};
export declare type ExecOptions = {
    resultMapper?: ResultMapper;
    fragmentMatcher?: FragmentMatcher;
};
export declare class LocalState<TCacheShape> {
    private cache;
    private client;
    private resolvers;
    private typeDefs;
    private fragmentMatcher;
    private firedInitializers;
    constructor({ cache, client, initializers, resolvers, typeDefs, fragmentMatcher, }: {
        cache: ApolloCache<TCacheShape>;
        client?: ApolloClient<TCacheShape>;
        initializers?: Initializers<TCacheShape> | Initializers<TCacheShape>[];
        resolvers?: Resolvers | Resolvers[];
        typeDefs?: string | string[] | DocumentNode | DocumentNode[];
        fragmentMatcher?: FragmentMatcher;
    });
    runInitializers(initializers: Initializers<TCacheShape> | Initializers<TCacheShape>[]): Promise<void[]>;
    runInitializersSync(initializers: Initializers<TCacheShape> | Initializers<TCacheShape>[]): void;
    addResolvers(resolvers: Resolvers | Resolvers[]): void;
    setResolvers(resolvers: Resolvers | Resolvers[]): void;
    getResolvers(): Resolvers | Resolvers[];
    runResolvers({ document, remoteResult, context, variables, onError, }: {
        document: DocumentNode | null;
        remoteResult?: ExecutionResult;
        context?: Record<string, any>;
        variables?: Record<string, any>;
        onError?: (error: any) => void;
    }): Promise<{
        errors?: ReadonlyArray<import("graphql").GraphQLError> | undefined;
        data?: import("graphql/execution/execute").ExecutionResultDataDefault | undefined;
    } | undefined>;
    setTypeDefs(typeDefs: string | string[] | DocumentNode | DocumentNode[]): void;
    getTypeDefs(): string | string[] | DocumentNode | DocumentNode[];
    setFragmentMatcher(fragmentMatcher: FragmentMatcher): void;
    getFragmentMatcher(): FragmentMatcher;
    clientQuery(document: DocumentNode): DocumentNode | null;
    serverQuery(document: DocumentNode): DocumentNode | null;
    prepareContext(context?: {}): {
        cache: ApolloCache<TCacheShape>;
        getCacheKey: (obj: {
            __typename: string;
            id: string | number;
        }) => any;
        schemas: object[];
    };
    addExportedVariables(document: DocumentNode, variables?: OperationVariables, context?: {}): Promise<{
        [x: string]: any;
    }>;
    resetInitializers(): void;
    private mergeInitializers;
    private processInitializers;
    private buildRootValueFromCache;
    private normalizeTypeDefs;
    private resolveDocument;
    private resolveSelectionSet;
    private resolveField;
    private resolveSubSelectedArray;
    private mergeIntoResults;
}
//# sourceMappingURL=LocalState.d.ts.map